from bpy.types import Operator
from bpy.props import StringProperty
import bpy
from . import trail_handler
from bpy.utils import register_classes_factory


def create_trail_material():
    from mathutils import Vector
    mat = bpy.data.materials.new("Trail Material")
    mat.blend_method = "BLEND"
    mat.shadow_method = "NONE"
    mat.use_nodes = True
    nt = mat.node_tree
    nt.nodes.clear()
    nodes = list(map(nt.nodes.new,
                     ['ShaderNodeBsdfTransparent', 'ShaderNodeSeparateXYZ', 'ShaderNodeUVMap', 'ShaderNodeEmission',
                      'ShaderNodeMath', 'ShaderNodeOutputMaterial', 'ShaderNodeMixShader', 'ShaderNodeMath']))
    for n, x in zip(nodes, [Vector((-160.0, 220.0)), Vector((-320.0, 400.0)), Vector((-500.0, 400.0)),
                            Vector((-160.0, 120.0)), Vector((-160.0, 400.0)), Vector((340.0, 280.0)),
                            Vector((180.0, 280.0)), Vector((0.0, 400.0))]):
        n.location = x
    for n1, s1, n2, s2 in [(6, 0, 5, 0), (3, 0, 6, 2), (0, 0, 6, 1), (2, 0, 1, 0), (4, 0, 7, 0), (1, 1, 4, 1),
                           (7, 0, 6, 0)]:
        nt.links.new(nodes[n1].outputs[s1], nodes[n2].inputs[s2])
    nodes[4].inputs[0].default_value = 1
    nodes[4].operation = 'SUBTRACT'
    nodes[7].inputs[1].default_value = 2
    nodes[7].operation = 'POWER'
    return mat


class OBJECT_OT_AddMeshTrailsMaterialPreset(Operator):
    """Add basic trail shader material to this mesh."""
    bl_label = "Add base trail material"
    bl_idname = 'meshtrails.add_preset_material'

    @classmethod
    def poll(cls, ctx):
        return ctx.object and ctx.object.type == 'MESH'

    def execute(self, context):
        mat = create_trail_material()
        if len(context.object.data.materials):
            context.object.data.materials[0] = mat
        else:
            context.object.data.materials.append(mat)
        return {'FINISHED'}


class OBJECT_OT_SelectObjByNameTrails(Operator):
    """Select this Object"""
    bl_label = "Select Object"
    bl_idname = 'meshtrails.trail_select_by_name'

    obj_name: StringProperty()

    def execute(self, context):
        try:
            ob = bpy.context.scene.objects[self.obj_name]
        except KeyError:
            return {'CANCELLED'}
        bpy.ops.object.select_all(action='DESELECT')
        bpy.context.view_layer.objects.active = ob
        ob.select_set(True)
        return {'FINISHED'}


class CURVE_OT_AddTrail(Operator):
    """ Add mesh trail to this curve """
    bl_idname = 'meshtrails.add_mesh_trail'
    bl_label = "Add Mesh Trail Object"

    @classmethod
    def poll(cls, context):
        return context.object.type == 'CURVE' and context.mode == 'OBJECT'

    def execute(self, context):
        ob = context.object
        me = bpy.data.meshes.new(f'{ob.name}.trail')
        me_ob = bpy.data.objects.new(f'{ob.name}.trail', me)
        me_ob.trail.target_object = ob
        try:
            context.collection.objects.link(me_ob)
        # Link to scene in the event the context collection is unusable
        except Exception:
            context.scene.collection.objects.link(me_ob)
        bpy.ops.object.select_all(action='DESELECT')
        context.view_layer.objects.active = me_ob
        me_ob.select_set(True)
        me.use_auto_smooth = True
        me.auto_smooth_angle = 3.14145/2

        mod = me_ob.modifiers.new("Subdivision Surface", 'SUBSURF')
        mod.levels = 1
        mod.render_levels = 1
        mod.boundary_smooth = 'PRESERVE_CORNERS'
        return {'FINISHED'}


class CURVE_OT_BakeTrail(Operator):
    """ Bake only this trail"""
    bl_idname = 'meshtrails.bake'
    bl_label = 'Bake Trail'

    @classmethod
    def poll(cls, context):
        return context.scene is not None and bpy.data.is_saved

    def execute(self, context):
        trail_handler.bake_single_trail(context.object)
        trail_handler.save_cache()
        return {'FINISHED'}


class CURVE_OT_BakeAllTrails(Operator):
    """ Bake all trails to cache """
    bl_idname = 'meshtrails.bake_all'
    bl_label = 'Bake All Trails'

    @classmethod
    def poll(cls, context):
        return context.scene is not None and bpy.data.is_saved

    def execute(self, context):
        trail_handler.bake_all_trails()
        trail_handler.save_cache()
        return {'FINISHED'}


class CURVE_OT_FreeTrail(Operator):
    """ Free cache for this trail """
    bl_idname = 'meshtrails.free'
    bl_label = 'Free Cache'

    @classmethod
    def poll(cls, context):
        return context.scene is not None and bpy.data.is_saved

    def execute(self, context):
        trail_handler.free_cache(context.object)
        return {'FINISHED'}


class CURVE_OT_FreeAllTrails(Operator):
    """ Free cache for this trail """
    bl_idname = 'meshtrails.free_all'
    bl_label = 'Free All Caches'

    @classmethod
    def poll(cls, context):
        return context.scene is not None and bpy.data.is_saved

    def execute(self, context):
        try:
            trail_handler.free_all()
        except FileNotFoundError as e:
            self.report({'WARNING'}, str(e))
        return {'FINISHED'}


class CURVE_OT_RefreshTrails(Operator):
    """ Reload trail handlers for all trails in scene """
    bl_idname = 'meshtrails.refresh'
    bl_label = 'Refresh Trails'

    @classmethod
    def poll(cls, context):
        return context.scene is not None

    def execute(self, context):
        trail_handler.load_handler(None)
        return {'FINISHED'}


_classes = [
    CURVE_OT_BakeAllTrails,
    CURVE_OT_BakeTrail,
    CURVE_OT_FreeAllTrails,
    CURVE_OT_FreeTrail,
    CURVE_OT_RefreshTrails,
    OBJECT_OT_AddMeshTrailsMaterialPreset,
    OBJECT_OT_SelectObjByNameTrails,
    CURVE_OT_AddTrail
]

register, unregister = register_classes_factory(_classes)
