from bpy.types import Panel
from .trail_utils import get_curve_data, speed_attr_enabled
from .trail_handler import is_cached, get_trail_info
import bpy
import os
from bpy.utils import register_classes_factory

script_file = os.path.realpath(__file__)
script_dir = os.path.dirname(script_file)
mat_lib_blend = os.path.abspath(os.path.join(script_dir, 'library.blend\\Material\\'))


class CURVE_PT_MeshTrails(Panel):
    bl_label = "Mesh Trails"
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "data"

    @classmethod
    def poll(cls, context):
        return context.object and context.curve

    def draw(self, context):
        layout = self.layout
        has_mesh = False
        col = layout.column(align=True)
        for obj in context.view_layer.objects:
            if obj.type != 'MESH':
                continue
            if obj.trail.target_object == context.object:
                has_mesh = True
                op = col.operator('meshtrails.trail_select_by_name', text=f"{obj.name}", icon='RESTRICT_SELECT_OFF')
                op.obj_name = obj.name

        if not has_mesh:
            layout.operator('meshtrails.add_mesh_trail')


class MESH_PT_MeshTrails(Panel):
    bl_label = "Mesh Trail"
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "data"
    bl_options = {'DEFAULT_CLOSED'}

    @classmethod
    def poll(cls, context):
        return context.mesh is not None

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        cached = is_cached(context.object)

        mesh = context.object
        trail = mesh.trail

        layout.active = True
        if trail.target_object:
            op = layout.operator('meshtrails.trail_select_by_name', text="Select Curve", icon='RESTRICT_SELECT_OFF')
            op.obj_name = trail.target_object.name

        col = layout.column()
        col.enabled = not cached
        col.prop(trail, "target_object")

        if not trail.target_object:
            return

        col.prop(trail, "space_object")

        # Only show the option in 3.0+
        if speed_attr_enabled():
            col.prop(trail, 'use_attribute', text="Calculate Speed")

        col.prop(trail, 'use_resolution')
        if not trail.use_resolution:
            col.prop(trail, 'resolution')
        col.prop(trail, 'length_frames')


        if not len(context.object.data.materials):
            col.operator('meshtrails.add_preset_material', text="Generate Simple Shader", icon='MATERIAL_DATA')

        warnings = [
            "Object not centred, trail may be offset from emitter." if
                any(val != 0.0 for val in list(context.object.location) + list(context.object.rotation_euler)) or
                any(val != 1.0 for val in list(context.object.scale)) else None,
            "Curve has unsupported modifiers, trail may not match curve shape." if
                trail.target_object and any(mod.type != "HOOK" for mod in trail.target_object.modifiers) else None,
            "Curve has multiple splines, only one spline will be used." if
                trail.target_object and len(trail.target_object.data.splines) > 1 else None,
            "Curve has no splines, no trail will be visible." if
                trail.target_object and len(trail.target_object.data.splines) == 0 else None,
            "Spline type unsupported, must be POLY or BEZIER." if
                trail.target_object and len(trail.target_object.data.splines) > 0 and
                trail.target_object.data.splines[0].type not in {"POLY", "BEZIER"} else None
        ]

        if any(warnings):
            box = layout.box()
            box.label(text="Warnings:", icon='ERROR')
            for warn in warnings:
                if warn:
                    box.label(text=warn)


class MTSubPanel(Panel):
    bl_space_type = 'PROPERTIES'
    bl_region_type = 'WINDOW'
    bl_context = "data"
    bl_options = {'DEFAULT_CLOSED'}
    bl_parent_id = "MESH_PT_MeshTrails"

    @classmethod
    def poll(cls, context):
        trail = context.object.trail
        return trail.target_object is not None


class MESH_PT_MeshTrailsWind(MTSubPanel):
    bl_label = "Wind"

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        cached = is_cached(context.object)
        layout.enabled = not cached

        mesh = context.object
        trail = mesh.trail

        layout.prop(trail, 'wind_mode')
        if trail.wind_mode == 'VECTOR':
            layout.prop(trail, 'wind_vector')
        if trail.wind_mode == 'OBJECT':
            layout.prop_search(trail, 'wind_object', context.scene, 'objects')
            layout.prop(trail, 'wind_object_axis')
        if trail.wind_mode != 'NONE':
            layout.prop(trail, 'wind_scale')


class MESH_PT_MeshTrailsTaper(MTSubPanel):
    bl_label = "Taper"

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True

        ob = context.object
        trail = ob.trail

        layout.prop(trail, 'taper_mode')
        if trail.taper_mode == 'LINEAR':
            layout.prop(trail, 'taper')
        if trail.taper_mode == 'CUSTOM':
            layout.prop(trail, 'taper')
            layout.template_curve_mapping(get_curve_data(trail), "map_curve")


class MESH_PT_MeshTrailsCache(MTSubPanel):
    bl_label = "Cache"

    def draw(self, context):
        layout = self.layout
        layout.use_property_split = True
        trail = context.object.trail
        cached = is_cached(context.object)

        row = layout.row()
        row.prop(trail, 'use_subframes')
        row.enabled = not cached

        row = layout.row()
        row.prop(context.scene.trail_settings, 'subframes')
        row.enabled = trail.use_subframes and not cached

        if not bpy.data.is_saved:
            layout.label(text="Caching disabled, file not saved.")
        t_info = get_trail_info(context.object)
        if t_info:
            layout.label(text=t_info)

        row = layout.row()
        row.prop(context.scene.trail_settings, 'cache_path')

        flow = layout.grid_flow(row_major=True, columns=0, even_columns=True, even_rows=False, align=False)

        col = flow.column()
        if cached:
            col.operator('meshtrails.free')
        else:
            col.operator('meshtrails.bake')

        col.operator('meshtrails.refresh')

        col = flow.column()
        col.operator('meshtrails.bake_all')
        col.operator('meshtrails.free_all')


_classes = [
    MESH_PT_MeshTrails,
    CURVE_PT_MeshTrails,
    MESH_PT_MeshTrailsWind,
    MESH_PT_MeshTrailsTaper,
    MESH_PT_MeshTrailsCache
]

register, unregister = register_classes_factory(_classes)
