from bpy.props import *
from bpy.types import PropertyGroup, Object, Scene, AddonPreferences
from . trail_handler import add_trail, remove_trail, update_trails
from . trail_utils import setup_taper_mod
from bpy.utils import register_classes_factory


wind_axes = [
    ('X_POS', 'X+', '', 0),
    ('X_NEG', 'X-', '', 1),
    ('Y_POS', 'Y+', '', 2),
    ('Y_NEG', 'Y-', '', 3),
    ('Z_POS', 'Z+', '', 4),
    ('Z_NEG', 'Z-', '', 5),
]


wind_modes = [
    ('NONE', 'None', 'No wind', 0),
    ('VECTOR', 'Vector', 'Wind defined by vector components', 1),
    ('OBJECT', 'Object', 'Wind controlled by target object', 2)
]


taper_modes = [
    ('NONE', 'None', 'No Taper', 0),
    ('LINEAR', 'Linear', 'Taper trail linearly', 1),
    ('CUSTOM', 'Custom', 'Taper with custom curve', 2)
]


def mapping_update(self, context):
    obj = self.id_data
    if self.target_object is None:
        remove_trail(obj)
    else:
        add_trail(obj)


def curve_poll(self, obj):
    return obj.type == 'CURVE'


def taper_update(self, context):
    setup_taper_mod(self.id_data.id_data)
    if context.screen and not context.screen.is_animation_playing:
        update_trails(context.view_layer.depsgraph)


def update_taper_mode(self, context):
    taper_update(self, context)


class SceneTrailSettings(PropertyGroup):
    subframes: IntProperty(
        name='Subframes (Global)',
        description="Visible after Bake only. Number of extra samples to take between each frame (Shared between all "
                    "trails in scene)",
        default=0,
        min=0,
        soft_max=15,
        options=set()
    )
    cache_path: StringProperty(
        name="Cache Path",
        description="File path for meshtrails cache",
        options=set(),
        subtype='FILE_PATH'
    )


class TrailSettings(PropertyGroup):

    target_object: PointerProperty(
        type=Object,
        name='Curve Object',
        description="Curve object for trail to follow",
        options=set(),
        poll=curve_poll,
        update=mapping_update
    )

    space_object: PointerProperty(
        type=Object,
        name='Space Object',
        description="Optionally calculate trail motion in this object's local space",
        options=set()
    )

    use_resolution: BoolProperty(
        name='Use Curve Resolution',
        description="Use Curve resolution as Trail resolution",
        default=True,
        options=set()
    )

    resolution: IntProperty(
        name='Resolution',
        description="Resolution of curve sample",
        default=10,
        min=2,
        soft_max=32,
        options=set()
    )

    use_subframes: BoolProperty(
        name='Use Subframes',
        description='Use sub-frame samples',
        default=True,
        options=set()
    )

    wind_mode: EnumProperty(
        items=wind_modes,
        name='Wind Mode',
        description='Add to previous trail positions for a wind/sliding effect'
    )

    wind_scale: FloatProperty(
        name='Wind scale',
        description='How far the trail should offset each frame',
        default=1.0
    )

    wind_vector: FloatVectorProperty(
        name='Wind Direction',
        description='Offset trail in this direction'
    )

    wind_object: PointerProperty(
        type=Object,
        name='Wind obj',
        description='Object to use to control wind direction'
    )

    wind_object_axis: EnumProperty(
        items=wind_axes,
        name='Axis',
        description='Axis of wind control object to use'
    )

    # Main trail properties

    length_frames: IntProperty(
        name='Length',
        description="length of trail, in frames",
        default=3,
        min=0,
        soft_max=32
    )

    taper_mode: EnumProperty(
        items=taper_modes,
        name='Taper',
        description='Taper the trail over its length',
        update=update_taper_mode
    )

    taper: FloatProperty(
        name='Factor',
        description="Relative taper amount of tail",
        default=1.0,
        min=0.0,
        soft_max=1.0,
        subtype='FACTOR',
        update=taper_update
    )

    use_attribute: BoolProperty(
        name="Create speed attribute",
        description="Add a \"trail_speed\" attribute to the mesh based on emitter motion",
        default=True,
        options=set()
    )


class AddonSettings(AddonPreferences):
    bl_idname = __package__

    enable_speed_attr: BoolProperty(
        name="EXPERIMENTAL: Enable Speed Attribute",
        description="This option is only available in 3.0. "
                    "Significantly higher chance of crashing during renders, use at your own risk",
        default=False
    )

    def draw(self, ctx):
        layout = self.layout
        layout.prop(self, 'enable_speed_attr')
        layout.label(text="Support:")
        row = layout.row()
        row.operator('wm.url_open', text='Report a bug', icon='ERROR').url = \
            'https://github.com/dillongoostudios/MeshTrailsSupport/issues/new?labels=bug&title=New+bug+report&body=Describe+your+issue+here'
        row.operator('wm.url_open', text="Support Forum", icon='HELP').url = \
            'https://github.com/dillongoostudios/MeshTrailsSupport/discussions'
        row.operator('wm.url_open', text='Contact Us', icon='QUESTION').url = \
            'mailto:support@dillongoostudios.com?subject=MeshTrails+Support'


_classes = [
    AddonSettings,
    SceneTrailSettings,
    TrailSettings,
]

_register, _unregister = register_classes_factory(_classes)


def register():
    _register()
    Object.trail = PointerProperty(type=TrailSettings)
    Scene.trail_settings = PointerProperty(type=SceneTrailSettings)


def unregister():
    _unregister()
    delattr(Object, 'trail')
    delattr(Scene, 'trail_settings')
