from bpy.app.handlers import persistent, frame_change_post, load_post, save_pre, save_post
from . trail_utils import *
import bpy
import os
import pickle

trails = {}
cache_scene = None

all_enabled = True


def update_trails(dg):
    for trail in trails.values():
        trail.update(dg)


@persistent
def load_handler(_):
    load_trails(bpy.context.scene)


def load_trails(scene):
    global cache_scene
    cache_scene = scene

    trails.clear()
    for obj in bpy.context.scene.objects:
        if obj.type == 'MESH':
            if obj.trail.target_object:
                trails[obj.name] = TrailManager(obj)
    load_cache()


@persistent
def frame_handler(scene, dg):
    if not all_enabled or scene is None:
        return
    if scene != cache_scene:
        load_trails(scene)
    if round(scene.frame_current) == frame_handler.frame:
        return

    # Undo can cause references to become out of date. Rebuild trail references if that is the case.
    if check_refresh():
        load_handler(None)

    frame_handler.frame = round(scene.frame_current)
    update_trails(dg)


frame_handler.frame = -1


def add_trail(obj):
    if obj.name not in trails:
        trails[obj.name] = TrailManager(obj)


def is_cached(obj):
    if obj.name not in trails:
        return False
    return trails[obj.name].cache.cached


def get_trail_info(obj):
    if obj.name not in trails:
        return None
    trail = trails[obj.name]
    if not trail.cache.cached:
        return None
    return f"{trail.cache.length()} frames in memory" #  ({trail.cache.total_size() / 1000}kiB)"


def free_all():
    for obj_name in trails:
        trails[obj_name].cache.cached = False
    os.remove(get_cache_path())


def free_cache(obj):
    if obj.name in trails:
        trails[obj.name].cache.cached = False
    try:
        data = pickle.load(open(get_cache_path(), 'rb'))
    except FileNotFoundError:
        return
    if obj.name in data:
        data.pop(obj.name)
    pickle.dump(data, open(get_cache_path(), 'wb'))


def remove_trail(obj):
    if obj.name in trails:
        trails.pop(obj.name)


def bake_single_trail(obj):
    global all_enabled
    load_handler(None)

    all_enabled = False
    add_trail(obj)
    ts = [trails[obj.name]]
    bake_trails(ts)
    all_enabled = True


def bake_all_trails():
    global all_enabled
    load_handler(None)

    all_enabled = False
    ts = [t for t in trails.values() if t.poll()]
    bake_trails(ts)
    all_enabled = True


def get_cache_path():
    trail_settings = bpy.context.scene.trail_settings
    fp = os.path.dirname(bpy.data.filepath)
    if trail_settings.cache_path == "":
        trail_settings.cache_path = f'//{bpy.path.display_name_from_filepath(bpy.data.filepath)}.cache'

    return bpy.path.abspath(trail_settings.cache_path)


def load_cache():
    if not bpy.data.is_saved:
        return
    path = get_cache_path()
    try:
        print(f"Loading MeshTrail cache from: {path}")
        data = pickle.load(open(path, 'rb'))
    except FileNotFoundError:
        print(f"No MeshTrails cache found!")
        return
    for obj_name in data:
        obj = bpy.data.objects.get(obj_name)
        if obj is None:
            continue
        if obj.type != 'MESH':
            continue
        print("  ", obj_name, "found in cache file")
        tm = TrailManager(obj)
        trails[obj_name] = tm
        tm.cache.from_pickle(data[obj_name])
        tm.cache.cached = True


def save_cache():
    if not bpy.data.is_saved:
        return
    # load_handler(None)
    data = {}
    for t in trails.values():
        if t.cache.cached:
            data[t.obj.name] = t.cache.tuple_points()
    pickle.dump(data, open(get_cache_path(), 'wb'))


def register():
    frame_change_post.append(frame_handler)
    load_post.append(load_handler)


def unregister():
    frame_change_post.remove(frame_handler)
    load_post.remove(load_handler)
