import math

from mathutils import Vector
from mathutils.geometry import interpolate_bezier
from collections import deque
import bmesh
import bpy


REFRESH_TRAILS = False
GRID_CREATE_OLD = bpy.app.version[0] < 3 and bpy.app.version[2] <= 2 and bpy.app.version[1] <= 90


def speed_attr_enabled():
    prefs = bpy.context.preferences.addons[__package__].preferences
    return bpy.app.version[0] >= 3 and prefs.enable_speed_attr


def check_refresh():
    global REFRESH_TRAILS
    if REFRESH_TRAILS:
        REFRESH_TRAILS = False
        return True
    return False


taper_mod_name = '_MeshTrailsTaperContainer'


def setup_taper_mod(obj):
    if obj.trail.taper_mode == 'CUSTOM':
        mod = obj.modifiers.get(taper_mod_name)
        if not mod or mod.type != 'VERTEX_WEIGHT_EDIT':
            if mod:
                obj.modifiers.remove(mod)
            mod = obj.modifiers.new(taper_mod_name, 'VERTEX_WEIGHT_EDIT')
            mod.show_viewport = False
            mod.show_render = False
            mod.show_expanded = False
            cu = mod.map_curve.curves[0]
            cu.points[0].location[1] = 1.0
            cu.points[1].location[1] = 0.0
            mod.map_curve.update()


def get_curve_data(trail):
    mod = trail.id_data.modifiers.get(taper_mod_name)
    if not mod or mod.type != 'VERTEX_WEIGHT_EDIT':
        if mod:
            trail.id_data.modifiers.remove(mod)
        mod = trail.id_data.modifiers.new(taper_mod_name, 'VERTEX_WEIGHT_EDIT')
        # Can't write ID data to the modifier since we're calling this from the UI, so we're stuck with defaults.
        # This will only happen if the taper mode was changed without running the update handler for some reason
    return mod


def set_grid_verts(mesh_obj, points, width=10, smooth=True, cyclic=True, subframes=1):
    mesh = mesh_obj.data
    use_attr = mesh_obj.trail.use_attribute and speed_attr_enabled()
    length = len(points) // max(width, 1)
    attr_name = 'trail_speed'

    if width < 2 or points is None or length <= 1:
        bm = bmesh.new()
        bm.to_mesh(mesh)
        bm.free()
        return

    attr = mesh.attributes.get(attr_name)

    # This only works for the case where the attribute is *enabled* for the first time. We can't remove the attribute
    # at render time.
    attr_changed = (attr is not None) != use_attr

    #  Case to reconstruct mesh
    if len(mesh.vertices) != len(points) or attr_changed:
        bm = bmesh.new(use_operators=True)
        bm.loops.layers.uv.verify()
        for v in bmesh.ops.create_grid(bm,
                                       x_segments=width if GRID_CREATE_OLD else width - 1,
                                       y_segments=length if GRID_CREATE_OLD else length - 1,
                                       size=1, calc_uvs=True)['verts']:
            v.co = points[v.index]
        if smooth:
            # Merging doubles dirties our indices, so disable it if we want to use them later (for attribute writing)
            if cyclic and not use_attr:
                cyclic_edge = [v for i, v in enumerate(bm.verts) if (i % width == 0 or i % width == width-1)]
                bmesh.ops.remove_doubles(bm, verts=list(cyclic_edge), dist=0.0001)
            for f in bm.faces:
                f.smooth = True
        try:
            bm.to_mesh(mesh)
        except ValueError:
            pass
        bm.free()

    #  Case to just set the vertex locations directly
    else:
        for v in mesh.vertices:
            v.co = points[v.index]

    if use_attr:
        # Get it again, since we may have cleared mesh data
        attr = mesh.attributes.get(attr_name)
        if attr is None or attr.data_type != 'FLOAT' or attr.domain != 'POINT':
            if attr is not None:
                mesh.attributes.remove(attr)
            attr = mesh.attributes.new(attr_name, 'FLOAT', 'POINT')
        # Loop through point pairs, set leading point attr to distance to next point
        for attr_data, cur_pos, next_pos in zip(attr.data, points, points[width:]):
            attr_data.value = (cur_pos - next_pos).length * subframes
        # Tail edge to 0
        for attr_data in attr.data[-width:]:
            attr_data.value = 0.0
    else:
        if attr is not None:
            # This causes GPU crashes mid-render. Just live with keeping the attribute.
            # mesh.attributes.remove(attr)
            pass


def mult_hook(mod, v):
    return mod.strength * (mod.object.matrix_world @ mod.matrix_inverse @ v) + \
           ((1 - mod.strength) * (mod.id_data.matrix_world @ v))


def sample_curve(obj, trail):
    curve = obj.data
    if curve is None or len(curve.splines) == 0:
        return []
    spline = curve.splines[0]

    if spline.type not in {'POLY', 'BEZIER'}:
        print("Mesh trail on invalid spline!")
        return []

    if spline.type == 'POLY':
        c = [obj.matrix_world @ p.co.xyz for p in spline.points]

        for mod in obj.modifiers:
            if mod.type == 'HOOK' and mod.object is not None:
                for index in mod.vertex_indices:
                    c[index] = mult_hook(mod, spline.points[index].co)
        return c

    points = spline.bezier_points

    # Copy original spline (just the bits we care about)
    mat = obj.matrix_world

    c = [mat @ p.co for p in points]
    l = [mat @ p.handle_left for p in points]
    r = [mat @ p.handle_right for p in points]

    c_o = [p.co for p in points]
    l_o = [p.handle_left for p in points]
    r_o = [p.handle_right for p in points]

    for mod in obj.modifiers:
        # Evaluate hooks
        if mod.type == 'HOOK' and mod.object is not None:
            for index in mod.vertex_indices:
                # Left handle
                if index % 3 == 0:
                    l[index // 3] = mult_hook(mod, l_o[index // 3])
                # co
                if index % 3 == 1:
                    c[index // 3] = mult_hook(mod, c_o[index // 3])
                # right handle
                if index % 3 == 2:
                    r[index // 3] = mult_hook(mod, r_o[index // 3])

    cyclic = spline.use_cyclic_u
    res = spline.resolution_u + 1 if trail.use_resolution else trail.resolution + 1
    p_last = -1

    out = []  # Returned list of points
    for i in range(len(c)):
        # Skip first iteration if it's not cyclic
        if i == 0 and not cyclic:
            p_last = i
            continue
        if i < len(c)-1:
            out += (interpolate_bezier(c[p_last], r[p_last], l[i], c[i], res))[:-1]
        else:
            out += (interpolate_bezier(c[p_last], r[p_last], l[i], c[i], res))
        p_last = i

    return out


def get_trail_resolution(obj):
    curve_obj = obj.trail.target_object
    if curve_obj is None:
        return 0
    if curve_obj.type != 'CURVE':
        return 0
    trail = obj.trail
    if len(curve_obj.data.splines) == 0:
        return 0
    spline = curve_obj.data.splines[0]
    if spline.type == 'POLY':
        return len(spline.points)
    res = spline.resolution_u if trail.use_resolution else trail.resolution

    res *= len(spline.bezier_points) - (1 if not spline.use_cyclic_u else 0)
    res += 1 if spline.use_cyclic_u else 1
    return res


def is_cyclic(obj):
    curve_obj = obj.trail.target_object
    if curve_obj is None:
        return False
    if curve_obj.type != 'CURVE':
        return False
    if len(curve_obj.data.splines) == 0:
        return False
    spline = curve_obj.data.splines[0]
    return spline.use_cyclic_u


vector_enums = {
        'X_POS': (1, 0, 0),
        'X_NEG': (-1, 0, 0),
        'Y_POS': (0, 1, 0),
        'Y_NEG': (0, -1, 0),
        'Z_POS': (0, 0, 1),
        'Z_NEG': (0, 0, -1),
}


def get_vector_enum(string):
    try:
        return Vector(vector_enums.get(string))
    except KeyError:
        return Vector((0, 0, 0))


def get_wind_vec(trail):
    if trail.wind_mode == 'OBJECT':
        if trail.wind_object is None:
            return Vector((0, 0, 0))
        target = trail.wind_object
        v = get_vector_enum(trail.wind_object_axis)
        wind = (target.matrix_world @ v) - target.matrix_world.decompose()[0]
    elif trail.wind_mode == 'VECTOR':
        wind = Vector(trail.wind_vector)
    else:
        wind = Vector((0, 0, 0))

    wind *= trail.wind_scale
    return wind


def get_verts(sequence):
    #  Just unpack
    return [vec for frame in sequence for vec in frame]


def get_mapping(trail):
    node = get_curve_data(trail)
    curve = node.map_curve.curves[0]

    if hasattr(node.map_curve, 'evaluate'):
        def calc(fac):
            return node.map_curve.evaluate(curve, fac)
        return calc
    else:
        return lambda fac: curve.evaluate(fac)


def map_range(value, from_min, from_max, to_min, to_max):
    # Figure out how 'wide' each range is
    from_range = from_max - from_min
    to_range = to_max - to_min

    # Convert the left range into a 0-1 range (float)
    value_scaled = float(value - from_min) / float(from_range)

    # Convert the 0-1 range into a value in the right range.
    return to_min + (value_scaled * to_range)


def get_verts_tapered(sequence, trail, n_sub=1):
    #  Unpack and taper at the same time to save iterating twice

    # Subframes extend into next frame by 1 subframe, so ignore last frame in sequence for taper length in that case
    la = (len(sequence) - (1 if len(sequence) >= 1 and len(sequence[-1]) != len(sequence[0]) else 0)) - (1 / n_sub)

    if trail.taper_mode == 'NONE' or la == 0:
        return [vec for frame in sequence for vec in frame]
    if trail.taper_mode == 'LINEAR':
        return [vec for i, frame in enumerate(sequence) for vec in
                taper_points_subframed(frame,
                                       [trail.taper * map_range(j, 0, n_sub, i / la, (i + 1) / la)
                                        for j in range(n_sub)])]
    if trail.taper_mode == 'CUSTOM':
        return [vec for i, frame in enumerate(sequence) for vec in
                taper_points_subframed(frame,
                                       [trail.taper * (1 - get_mapping(trail)(map_range(
                                           j, 0, n_sub, i / la, (i+1) / la
                                       ))) for j in range(n_sub)])]


def get_mid_point(points, midpoint=0.5):
    if len(points) < 2:
        return Vector((0, 0, 0))
    return points[0].lerp(points[-1], midpoint)


def get_bounds_centre(points):
    min_ = None
    max_ = None
    for v in points:
        if min_ is None:
            min_ = list(v.to_tuple())
        if max_ is None:
            max_ = list(v.to_tuple())
        for i in range(0, 3):
            if v[i] < min_[i]:
                min_[i] = v[i]
            if v[i] > max_[i]:
                max_[i] = v[i]

    corner_low = Vector(tuple(min_))
    corner_high = Vector(tuple(max_))
    if len(points) < 2 or min_ is None or max_ is None:
        return Vector((0, 0, 0))
    return corner_low.lerp(corner_high, 0.5)


def scale_points(points, centre, factor):
    return [p.lerp(centre, factor) for p in points]


def taper_points(points, factor):
    return scale_points(points, get_bounds_centre(points), factor)


def chunks(data, n):
    n = max(1, n)
    return (data[i:i + n] for i in range(0, len(data), n))


def taper_points_subframed(points, factors):
    out = []
    nf = len(factors)
    for i, c in enumerate(chunks(points, len(points) // nf)):
        if i >= nf:
            break
        out += scale_points(c, get_bounds_centre(c), factors[i])
    return out


def eval_trail_current_frame(obj, previous, space_object=None):
    trail = obj.trail
    target_obj = trail.target_object
    if target_obj.type != 'CURVE':
        print("Trail on non-curve object")
        return []

    frames = deque(previous)

    while len(frames) > trail.length_frames:
        frames.pop()

    wind = get_wind_vec(trail)
    if wind.length_squared > 0.00001:  # Skip if not required
        # Convert wind vector to local space if needed.
        if space_object:
            wind = space_object.matrix_world.to_3x3().inverted() @ wind
        for i, frame in enumerate(frames):
            frames[i] = [p + wind for p in frame]

    if space_object:
        frames.appendleft(to_obj_space(space_object, sample_curve(target_obj, trail)))
    else:
        frames.appendleft(sample_curve(target_obj, trail))

    return list(frames)


def eval_trail_from_subframes(obj, previous, subframes):
    trail = obj.trail
    target_obj = trail.target_object
    if target_obj.type != 'CURVE':
        print("Trail on non-curve object")
        return []

    frames = deque(previous)

    # Shift back by one frame so subframes fill *up to* next frame.
    if trail.length_frames == 0:
        return []

    while len(frames) > trail.length_frames:
        frames.pop()

    # Grab last part of tailing frame
    if len(frames) == trail.length_frames:
        frames[-1] = frames[-1][:len(subframes[0])]

    wind = get_wind_vec(trail)
    if wind.length_squared > 0.0:  # Skip if we don't need it
        for i, frame in enumerate(frames):
            frames[i] = [p + wind for p in frame]

    n = len(subframes)

    if wind:
        for i, subf in enumerate(subframes):
            subframes[i] = [p + (wind * ((n - i - 1) * (1.0/n))) for p in subf]

    subframes.reverse()

    points = [p for f in subframes for p in f]
    frames.appendleft(points)
    return list(frames)


def to_obj_space(obj, points):
    mat = obj.matrix_world.inverted()
    return [mat @ p for p in points]


def from_obj_space(obj, points):
    mat = obj.matrix_world
    return [mat @ p for p in points]


def get_points_subframed(obj):
    trail = obj.trail
    target_obj = trail.target_object

    points = sample_curve(target_obj, trail)
    if trail.space_object:
        return to_obj_space(trail.space_object, points)
    return sample_curve(target_obj, trail)


def bake_subframes(scene, trails, frame, count):

    subframes = {}
    for t in trails:
        if t.use_subframes():
            subframes[t] = []

    if len(subframes) == 0:
        return

    offset = 1.0 / count
    cur_offset = offset
    for i in range(count):
        scene.frame_set(frame-1, subframe=cur_offset)
        for t in subframes:
            subframes[t].append(get_points_subframed(t.obj))
        cur_offset += offset

    for t in subframes:
        t.cache.update_subframes(subframes[t])


def bake_trails(trails):
    print("Baking trails...")
    scn = bpy.context.scene
    f_start = scn.frame_start
    f_end = scn.frame_end
    f_cur = scn.frame_current

    # 0 subframes = 1 frame
    count = scn.trail_settings.subframes + 1

    wm = bpy.context.window_manager
    wm.progress_begin(f_start, f_end+1)
    for t in trails:
        t.cache.clear(f_start)
        if t.use_subframes():
            t.cache.subframes = count
        else:
            t.cache.subframes = 1

    for f in range(f_start, f_end+1):
        if f == f_start:
            bake_subframes(scn, trails, f, 1)
        else:
            bake_subframes(scn, trails, f, count)

        # Just in case something goes wrong
        bpy.context.scene.frame_set(f)
        for t in trails:
            if not t.use_subframes():
                t.cache.update()

        wm.progress_update(f)

    scn.frame_set(f_cur)
    for t in trails:
        t.cache.cached = True
    wm.progress_end()
    print("Trail baking done!")


class RealtimeTrail:
    res = 0

    def __init__(self, obj):
        self.obj = obj
        self.points = []
        self.res = get_trail_resolution(obj)

    def get_points(self):
        new_res = get_trail_resolution(self.obj)
        if new_res == 0:
            return []
        if new_res != self.res:
            self.points.clear()
            self.res = new_res

        self.points = eval_trail_current_frame(self.obj, self.points, self.obj.trail.space_object)
        points = get_verts_tapered(self.points, self.obj.trail, 1)

        if self.obj.trail.space_object:
            return from_obj_space(self.obj.trail.space_object, points)
        return points


def get_scene_frame(scene):
    frame = bpy.context.scene.frame_float
    if scene.render.engine == 'BLENDER_EEVEE' and scene.eevee.use_motion_blur:
        return {
            "START": math.floor(frame),
            "END": math.ceil(frame),
            "CENTER": round(frame)
        }.get(scene.eevee.motion_blur_position)
    return math.floor(frame)


class CachedTrail:
    res = 0
    start = 0
    subframes = 1
    cached = False

    def __init__(self, obj):
        self.obj = obj
        self.frames = []

    def length(self):
        return len(self.frames)

    def total_size(self):
        # Include list overhead in frame size calculation
        # sizeof(frame) = len(frame) * (sizeof(*frame[0]) + sizeof(frame[0])) + sizeof([])
        return sum([len(frame) * (376 + 8) + 56 for frame in self.frames])

    def get_points(self):
        frame = get_scene_frame(bpy.context.scene)
        if not self.start <= frame < self.start + len(self.frames):
            return []
        points = get_verts_tapered(self.frames[frame - self.start], self.obj.trail, self.subframes)
        if self.obj.trail.space_object:
            return from_obj_space(self.obj.trail.space_object, points)
        return points

    def clear(self, start):
        self.frames.clear()
        self.start = start
        self.res = get_trail_resolution(self.obj)

    def update(self):
        points = eval_trail_current_frame(self.obj, self.frames[-1].copy() if len(self.frames) else [], self.obj.trail.space_object)
        self.frames.append(points)

    def update_subframes(self, subframes):
        points = eval_trail_from_subframes(self.obj, self.frames[-1].copy() if len(self.frames) else [], subframes)
        self.frames.append(points)

    def tuple_points(self):
        return [[[vec.to_tuple() for vec in row] for row in frame] for frame in self.frames], \
               self.start, \
               self.res, \
               self.subframes

    def from_pickle(self, data):
        self.frames = [[[Vector(tup) for tup in row] for row in frame] for frame in data[0]]
        self.start = data[1]
        self.res = data[2]
        self.subframes = data[3]


export_prop_name = "_mesh_trail_export"


# Add a '#frame' driver to the mesh data to hint to the alembic exporter that the mesh is animated
def init_export_driver(obj: bpy.types.Object):
    data = obj.data
    data[export_prop_name] = 0.0
    drv: bpy.types.Driver = data.driver_add(f'["{export_prop_name}"]').driver
    drv.type = 'SCRIPTED'
    drv.expression = 'frame'


class TrailManager:
    def __init__(self, obj):
        self.obj = obj
        self.preview = RealtimeTrail(obj)
        self.cache = CachedTrail(obj)
        init_export_driver(obj)

    def use_subframes(self):
        return self.obj.trail.use_subframes

    def poll(self):
        try:
            _ = self.obj.trail
            _ = self.obj.trail.target_object
        except ReferenceError as e:
            print(e)
            return False
        return True

    def update(self, dg):
        try:
            _ = self.obj.trail
        except ReferenceError:
            # New Undo system seems to break this if trail was not present on file save... but works otherwise?
            global REFRESH_TRAILS
            REFRESH_TRAILS = True
            return
        # Skip only if in viewport mode. Rendered mode may not be visible on frame change, but view layers can change
        if dg.mode != "RENDER" and not self.obj.visible_get():
            return

        if not self.cache.cached:
            points = self.preview.get_points()
            set_grid_verts(self.obj, points, self.preview.res, cyclic=is_cyclic(self.obj))
            return
        points = self.cache.get_points()
        set_grid_verts(self.obj, points, self.cache.res, cyclic=is_cyclic(self.obj), subframes=self.cache.subframes)
        return

